"""
Client for communicating with the Enterprise API.
"""

import json
from collections import OrderedDict
from logging import getLogger
from urllib.parse import urljoin

from requests.exceptions import (  # pylint: disable=redefined-builtin
    ConnectionError,
    HTTPError,
    RequestException,
    Timeout,
)

from django.conf import settings

from enterprise import utils
from enterprise.api_client.client import NoAuthAPIClient, UserAPIClient

LOGGER = getLogger(__name__)


class EnterpriseCatalogApiClient(UserAPIClient):
    """
    The API client to make calls to the Enterprise Catalog API.
    """

    API_BASE_URL = urljoin(f"{settings.ENTERPRISE_CATALOG_INTERNAL_ROOT_URL}/", "api/v1/")
    ENTERPRISE_CATALOG_ENDPOINT = 'enterprise-catalogs'
    GET_CONTENT_METADATA_ENDPOINT = ENTERPRISE_CATALOG_ENDPOINT + '/{}/get_content_metadata'
    REFRESH_CATALOG_ENDPOINT = ENTERPRISE_CATALOG_ENDPOINT + '/{}/refresh_metadata'
    CATALOG_DIFF_ENDPOINT = ENTERPRISE_CATALOG_ENDPOINT + '/{}/generate_diff'
    ENTERPRISE_CUSTOMER_ENDPOINT = 'enterprise-customer'
    CUSTOMER_CONTENT_METADATA_IDENTIFIER_ENDPOINT = ENTERPRISE_CUSTOMER_ENDPOINT + \
        "/{}/content-metadata/" + "{}"
    CONTENT_METADATA_IDENTIFIER_ENDPOINT = "content-metadata/"
    CATALOG_QUERIES_ENDPOINT = 'catalog-queries'
    GET_CONTENT_FILTER_HASH_ENDPOINT = CATALOG_QUERIES_ENDPOINT + '/get_content_filter_hash'
    GET_QUERY_BY_HASH_ENDPOINT = CATALOG_QUERIES_ENDPOINT + '/get_query_by_hash?hash={}'
    APPEND_SLASH = True
    GET_CONTENT_METADATA_PAGE_SIZE = getattr(settings, 'ENTERPRISE_CATALOG_GET_CONTENT_METADATA_PAGE_SIZE', 50)

    def __init__(self, user=None):
        user = user if user else utils.get_enterprise_worker_user()
        super().__init__(user)

    @UserAPIClient.refresh_token
    def create_enterprise_catalog(
            self,
            catalog_uuid,
            enterprise_id,
            enterprise_name,
            title,
            content_filter,
            enabled_course_modes,
            publish_audit_enrollment_urls,
            catalog_query_uuid,
            query_title=None,
            include_exec_ed_2u_courses=False,
    ):
        """
        Creates an enterprise catalog.
        """
        api_url = self.get_api_url(f"{self.ENTERPRISE_CATALOG_ENDPOINT}")
        post_data = {
            'uuid': catalog_uuid,
            'title': title,
            'enterprise_customer': enterprise_id,
            'enterprise_customer_name': enterprise_name,
            'content_filter': content_filter,
            'enabled_course_modes': enabled_course_modes,
            'publish_audit_enrollment_urls': json.dumps(publish_audit_enrollment_urls),
            'catalog_query_uuid': catalog_query_uuid,
            'query_title': query_title,
            'include_exec_ed_2u_courses': include_exec_ed_2u_courses,
        }
        try:
            LOGGER.info(
                'Creating Enterprise Catalog %s in the Enterprise Catalog Service with params: %s',
                catalog_uuid,
                json.dumps(post_data)
            )
            response = self.client.post(api_url, json=post_data)
            response.raise_for_status()
            return response.json()
        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to create EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                catalog_uuid, str(exc)
            )
            return {}

    @UserAPIClient.refresh_token
    def get_enterprise_catalog(self, catalog_uuid, should_raise_exception=True):
        """
        Gets an enterprise catalog.

        Arguments:
            catalog_uuid (uuid): The uuid of an EnterpriseCatalog instance
            should_raise_exception (bool): Whether an exception should be logged if
                a catalog does not yet exist in enterprise-catalog. Defaults to True.

        Returns:
            dict: a dictionary representing an enterprise catalog
        """
        api_url = self.get_api_url(f"{self.ENTERPRISE_CATALOG_ENDPOINT}/{catalog_uuid}")
        try:
            response = self.client.get(api_url)
            response.raise_for_status()
            return response.json()
        except (RequestException, ConnectionError, Timeout) as exc:
            if should_raise_exception:
                LOGGER.exception(
                    'Failed to get EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                    catalog_uuid, str(exc)
                )
            return {}

    @UserAPIClient.refresh_token
    def get_enterprise_catalog_by_hash(self, catalog_query_hash, should_raise_exception=True):
        """
        Gets an enterprise catalog by its content query hash.

        Arguments:
            catalog_query_hash (str): The hash of an EnterpriseCatalog instance's query
            should_raise_exception (bool): Whether an exception should be logged if
                a catalog does not yet exist in enterprise-catalog. Defaults to True.

        Returns:
            dict: a dictionary representing an enterprise catalog
        """
        api_url = self.get_api_url(self.GET_QUERY_BY_HASH_ENDPOINT.format(catalog_query_hash), append_slash=False)
        try:
            response = self.client.get(api_url)
            response.raise_for_status()
            return response.json()
        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to get EnterpriseCustomer Catalog by hash [%s] in enterprise-catalog due to: [%s]',
                catalog_query_hash, str(exc)
            )
            if should_raise_exception:
                raise
            return {}

    @UserAPIClient.refresh_token
    def get_catalog_query_hash(self, catalog_query, should_raise_exception=True):
        """
        Gets hash for content query.

        Arguments:
            catalog_query (obj): content query json
            should_raise_exception (bool): Whether an exception should be logged if
                a catalog does not yet exist in enterprise-catalog. Defaults to True.

        Returns:
            dict: a dictionary representing an enterprise catalog
        """
        api_url = self.get_api_url(self.GET_CONTENT_FILTER_HASH_ENDPOINT, append_slash=False)
        try:
            response = self.client.get(api_url, data=catalog_query)
            response.raise_for_status()
            return response.json()
        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to get catalog query hash for "[%s]" due to: [%s]',
                catalog_query, str(exc)
            )
            if should_raise_exception:
                raise
            return {}

    @UserAPIClient.refresh_token
    def get_catalog_diff(self, enterprise_customer_catalog, content_keys, should_raise_exception=True):
        """
        Gets the representational difference between a list of course keys and the current state of content under an
        enterprise catalog. This difference is returned as three buckets of data: `items_not_found`,
        `items_not_included` and `items_found`.

        Arguments:
            enterprise_customer_catalog (EnterpriseCustomerCatalog): The catalog object whose content is being diffed.
            content_keys (list): List of string content keys
            should_raise_exception (Bool): Optional param for whether or not api response exceptions should be raised.

        Returns:
            items_to_create (list): dictionaries of content_keys to create
            items_to_delete (list): dictionaries of content_keys to delete
            items_found (list): dictionaries of content_keys and date_updated datetimes of content to update
        """
        catalog_uuid = enterprise_customer_catalog.uuid
        api_url = self.get_api_url(self.CATALOG_DIFF_ENDPOINT.format(catalog_uuid))
        body = {'content_keys': content_keys}

        items_to_delete = []
        items_to_create = []
        items_found = []
        try:
            response = self.client.post(api_url, json=body)
            response.raise_for_status()
            results = response.json()
            items_to_delete = results.get('items_not_found')
            items_to_create = results.get('items_not_included')
            items_found = results.get('items_found')

        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to get EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                catalog_uuid, str(exc)
            )
            if should_raise_exception:
                raise

        return items_to_create, items_to_delete, items_found

    @staticmethod
    def get_content_metadata_url(uuid):
        """
        Get the url for the preview information for an enterprise catalog
        """
        return EnterpriseCatalogApiClient.API_BASE_URL + \
            EnterpriseCatalogApiClient.GET_CONTENT_METADATA_ENDPOINT.format(uuid)

    @UserAPIClient.refresh_token
    def update_enterprise_catalog(self, catalog_uuid, **kwargs):
        """
        Updates an enterprise catalog.
        """
        api_url = self.get_api_url(f"{self.ENTERPRISE_CATALOG_ENDPOINT}/{catalog_uuid}")
        try:
            LOGGER.info(
                'Updating Enterprise Catalog %s in the Enterprise Catalog Service with params: %s',
                catalog_uuid,
                json.dumps(kwargs)
            )
            response = self.client.put(api_url, json=kwargs)
            response.raise_for_status()
            return response.json()
        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                'Failed to update EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                catalog_uuid, str(exc)
            )
            return {}

    @UserAPIClient.refresh_token
    def delete_enterprise_catalog(self, catalog_uuid):
        """
        Deletes an enterprise catalog.
        """
        api_url = self.get_api_url(f"{self.ENTERPRISE_CATALOG_ENDPOINT}/{catalog_uuid}")
        try:
            response = self.client.delete(api_url)
            response.raise_for_status()
            return 200 <= response.status_code <= 299
        except (RequestException, ConnectionError, Timeout) as exc:
            if exc.response is not None and exc.response.status_code == 404:
                LOGGER.warning(
                    'Deleted EnterpriseCustomerCatalog [%s] that was not in enterprise-catalog',
                    catalog_uuid
                )
            else:
                LOGGER.exception(
                    'Failed to delete EnterpriseCustomer Catalog [%s] in enterprise-catalog due to: [%s]',
                    catalog_uuid, str(exc)
                )
            return {}

    def traverse_get_content_metadata(self, api_url, query, catalog_uuid):
        """
        Helper method to traverse over a paginated response from the enterprise-catalog service's `get_content_metadata`
        endpoint.
        """
        content_metadata = OrderedDict()
        try:
            response = self.client.get(api_url, params=query)
            response.raise_for_status()
            for item in utils.traverse_pagination(response.json(), self.client, api_url):
                content_id = utils.get_content_metadata_item_id(item)
                content_metadata[content_id] = item
        except (RequestException, ConnectionError, Timeout):
            LOGGER.exception('Failed to get content metadata for Catalog %s in enterprise-catalog', catalog_uuid)
            raise

        return content_metadata

    @UserAPIClient.refresh_token
    def get_content_metadata(self, enterprise_customer, enterprise_catalogs=None, content_keys_filter=None):
        """
        Return all content metadata contained in the catalogs associated with the EnterpriseCustomer.

        Arguments:
            enterprise_customer (EnterpriseCustomer): The EnterpriseCustomer to return content metadata for.
            enterprise_catalogs (EnterpriseCustomerCatalog): Optional list of EnterpriseCustomerCatalog objects.
            content_keys_filter (List): List of content keys to filter by in the content metadata endpoint

        Returns:
            list: List of dicts containing content metadata.
        """
        content_metadata = OrderedDict()
        enterprise_customer_catalogs = enterprise_catalogs or enterprise_customer.enterprise_customer_catalogs.all()
        for enterprise_customer_catalog in enterprise_customer_catalogs:
            catalog_uuid = enterprise_customer_catalog.uuid
            api_url = self.get_api_url(self.GET_CONTENT_METADATA_ENDPOINT.format(catalog_uuid))
            # If content keys filter exists then chunk up the keys into reasonable request sizes
            if content_keys_filter:
                chunked_keys_filter = utils.batch(
                    content_keys_filter,
                    self.GET_CONTENT_METADATA_PAGE_SIZE
                )
                # A chunk can be larger than the page size so traverse pagination for each individual chunk
                for chunk in chunked_keys_filter:
                    query = {'page_size': self.GET_CONTENT_METADATA_PAGE_SIZE, 'content_keys': chunk}
                    content_metadata.update(self.traverse_get_content_metadata(api_url, query, catalog_uuid))
            # Traverse pagination for the get all content response without filters
            else:
                query = {'page_size': self.GET_CONTENT_METADATA_PAGE_SIZE}
                content_metadata.update(self.traverse_get_content_metadata(api_url, query, catalog_uuid))

        return list(content_metadata.values())

    @UserAPIClient.refresh_token
    def get_catalog_content_count(self, catalog_uuid):
        """
        Gets the content metadata count for a catalog from the first page of results
        """
        api_url = self.get_api_url(self.GET_CONTENT_METADATA_ENDPOINT.format(catalog_uuid))
        response = self.client.get(api_url)
        response.raise_for_status()
        return response.json()['count']

    @UserAPIClient.refresh_token
    def refresh_catalogs(self, enterprise_catalogs):
        """
        Kicks off async tasks to refresh catalogs so recent changes will populate to production without needing to wait
        for the daily jobs to run

        Arguments:
            enterprise_catalogs (EnterpriseCustomerCatalog): List of EnterpriseCustomerCatalog objects to refresh

        Returns:
            Dict: Dict of async task ids for each catalog id
        """
        refreshed_catalogs = {}
        failed_to_refresh_catalogs = []
        for enterprise_customer_catalog in enterprise_catalogs:
            catalog_uuid = enterprise_customer_catalog.uuid
            api_url = self.get_api_url(self.REFRESH_CATALOG_ENDPOINT.format(catalog_uuid))
            try:
                response = self.client.post(api_url)
                response.raise_for_status()
                refreshed_catalogs[catalog_uuid] = response.json()['async_task_id']
            except (RequestException, ConnectionError, Timeout):
                LOGGER.exception(
                    'Failed to refresh catalog data for catalog [%s] in enterprise-catalog.', catalog_uuid
                )
                failed_to_refresh_catalogs.append(catalog_uuid)

        return refreshed_catalogs, failed_to_refresh_catalogs

    @UserAPIClient.refresh_token
    def contains_content_items(self, catalog_uuid, content_ids):
        """
        Checks whether an enterprise catalog contains the given content.

        The enterprise catalog endpoint does not differentiate between course_run_ids and program_uuids so they can
        be used interchangeably. The two query parameters are left in for backwards compatability with edx-enterprise.
        """
        query_params = {'course_run_ids': content_ids}
        api_url = self.get_api_url(f"{self.ENTERPRISE_CATALOG_ENDPOINT}/{catalog_uuid}/contains_content_items")
        response = self.client.get(api_url, params=query_params)
        response.raise_for_status()
        return response.json()['contains_content_items']

    @UserAPIClient.refresh_token
    def enterprise_contains_content_items(self, enterprise_uuid, content_ids):
        """
        Checks whether an enterprise customer has any catalogs that contain the provided content ids.

        The endpoint does not differentiate between course_run_ids and program_uuids so they can be used
        interchangeably. The two query parameters are left in for backwards compatability with edx-enterprise.
        """
        query_params = {
            'course_run_ids': content_ids,
            'get_catalogs_containing_specified_content_ids': True,
        }
        api_url = self.get_api_url(f"{self.ENTERPRISE_CUSTOMER_ENDPOINT}/{enterprise_uuid}/contains_content_items")
        response = self.client.get(api_url, params=query_params)
        response.raise_for_status()
        return response.json()

    @UserAPIClient.refresh_token
    def get_customer_content_metadata_content_identifier(self, enterprise_uuid, content_id):
        """
        Return all content metadata contained in the catalogs associated with the
        given EnterpriseCustomer and content_id.
        """
        try:
            api_url = self.get_api_url(
                f"{self.CUSTOMER_CONTENT_METADATA_IDENTIFIER_ENDPOINT.format(enterprise_uuid, content_id)}"
            )
            response = self.client.get(api_url)
            response.raise_for_status()
            return response.json()
        except HTTPError as exc:
            LOGGER.exception(
                "No matching content found in catalog for customer: [%s] or content_id: [%s], Error: %s",
                enterprise_uuid,
                content_id,
                str(exc),
            )
            return {}
        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                (
                    "Exception raised in "
                    "EnterpriseCatalogApiClient::get_customer_content_metadata_content_identifier: [%s]"
                ),
                str(exc),
            )
            return {}

    @UserAPIClient.refresh_token
    def get_content_metadata_content_identifier(self, content_id, coerce_to_parent_course=False):
        """
        Return the content metadata for the content_id.
        """
        try:
            api_url = self.get_api_url(self.CONTENT_METADATA_IDENTIFIER_ENDPOINT)
            query_params = {}
            if coerce_to_parent_course:
                query_params["coerce_to_parent_course"] = True
            response = self.client.get(
                api_url + content_id,
                params=query_params,
            )
            response.raise_for_status()
            return response.json()
        except HTTPError as exc:
            LOGGER.exception(
                "No matching content found for content_id: [%s], Error: %s",
                content_id,
                str(exc),
            )
            return {}
        except (RequestException, ConnectionError, Timeout) as exc:
            LOGGER.exception(
                "Exception raised in EnterpriseCatalogApiClient::get_content_metadata_content_identifier: [%s]",
                str(exc),
            )
            return {}


class NoAuthEnterpriseCatalogClient(NoAuthAPIClient):
    """
    The enterprise catalog API client to make calls to the discovery service.
    """

    API_BASE_URL = settings.ENTERPRISE_CATALOG_INTERNAL_ROOT_URL
    APPEND_SLASH = False



# Generated by Django 4.2.15 on 2024-08-26 07:03

from django.db import migrations
import fernet_fields.fields


class Migration(migrations.Migration):

    dependencies = [
        ('canvas', '0035_canvaslearnerassessmentdatatransmissionaudit_is_transmitted_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='canvasenterprisecustomerconfiguration',
            name='decrypted_client_id',
            field=fernet_fields.fields.EncryptedCharField(blank=True, default='', help_text='The encrypted API Client ID provided to edX by the enterprise customer to be used to make API calls to Canvas on behalf of the customer. It will be encrypted when stored in the database.', max_length=255, null=True, verbose_name='Encrypted API Client ID'),
        ),
        migrations.AddField(
            model_name='canvasenterprisecustomerconfiguration',
            name='decrypted_client_secret',
            field=fernet_fields.fields.EncryptedCharField(blank=True, default='', help_text='The encrypted API Client Secret provided to edX by the enterprise customer to be used to make  API calls to Canvas on behalf of the customer. It will be encrypted when stored in the database.', max_length=255, null=True, verbose_name='Encrypted API Client Secret'),
        ),
    ]



"""
Views for the ``enterprise-course-enrollment`` API endpoint.
"""
from django_filters.rest_framework import DjangoFilterBackend
from edx_rest_framework_extensions.paginators import DefaultPagination
from rest_framework import filters, permissions
from rest_framework.decorators import action
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST

from django.core.paginator import Paginator
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property

from enterprise import models
from enterprise.api.filters import EnterpriseCourseEnrollmentFilterBackend
from enterprise.api.utils import CourseRunProgressStatuses  # pylint: disable=cyclic-import
from enterprise.api.v1 import serializers
from enterprise.api.v1.views.base_views import EnterpriseReadWriteModelViewSet

try:
    from common.djangoapps.util.query import read_replica_or_default
except ImportError:
    def read_replica_or_default():
        return None
try:
    from openedx.core.djangoapps.content.course_overviews.api import get_course_overviews
except ImportError:
    get_course_overviews = None


class PaginatorWithOptimizedCount(Paginator):
    """
    Django < 4.2 ORM doesn't strip unused annotations from count queries.

    For example, if we execute this query:

        Book.objects.annotate(Count('chapters')).count()

    it will generate SQL like this:

        SELECT COUNT(*) FROM (SELECT COUNT(...), ... FROM ...) subquery

    This isn't optimal on its own, but it's not a big deal. However, this
    becomes problematic when annotations use subqueries, because it's terribly
    inefficient to execute the subquery for every row in the outer query.

    This class overrides the count() method of Django's Paginator to strip
    unused annotations from the query by requesting only the primary key
    instead of all fields.

    This is a temporary workaround until Django is updated to 4.2, which will
    include a fix for this issue.

    See https://code.djangoproject.com/ticket/32169 for more details.

    TODO: remove this class once Django is updated to 4.2 or higher.
    """
    @cached_property
    def count(self):
        return self.object_list.values("pk").count()


class EnterpriseCourseEnrollmentPagination(DefaultPagination):
    django_paginator_class = PaginatorWithOptimizedCount


class EnterpriseCourseEnrollmentViewSet(EnterpriseReadWriteModelViewSet):
    """
    API views for the ``enterprise-course-enrollment`` API endpoint.
    """

    queryset = models.EnterpriseCourseEnrollment.with_additional_fields.all()
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend, EnterpriseCourseEnrollmentFilterBackend)

    USER_ID_FILTER = 'enterprise_customer_user__user_id'
    FIELDS = (
        'enterprise_customer_user', 'course_id'
    )
    filterset_fields = FIELDS
    ordering_fields = FIELDS

    pagination_class = EnterpriseCourseEnrollmentPagination

    def get_queryset(self):
        queryset = super().get_queryset()
        if self.request.method == 'GET':
            queryset = queryset.using(read_replica_or_default())
        return queryset

    def get_serializer_class(self):
        """
        Use a special serializer for any requests that aren't read-only.
        """
        if self.request.method in ('GET',):
            return serializers.EnterpriseCourseEnrollmentWithAdditionalFieldsReadOnlySerializer
        return serializers.EnterpriseCourseEnrollmentWriteSerializer


class EnterpriseCourseEnrollmentAdminPagination(PageNumberPagination):
    """
    Custom pagination class for Enterprise Course Enrollment API.
    """
    page_size = 10
    page_size_query_param = 'page_size'
    max_page_size = 100


class EnterpriseCourseEnrollmentAdminViewSet(EnterpriseReadWriteModelViewSet):
    """
    API views for the ``enterprise-course-enrollment-admin`` API endpoint.
    """

    queryset = models.EnterpriseCourseEnrollment.with_additional_fields.all()
    serializer_class = serializers.EnterpriseCourseEnrollmentAdminViewSerializer
    permission_classes = (permissions.IsAuthenticated, permissions.IsAdminUser)
    pagination_class = EnterpriseCourseEnrollmentAdminPagination

    @action(detail=False, methods=['get'])
    def get_enterprise_course_enrollments(self, request):
        """
        Endpoint to get enrollments for a learner by `lms_user_id` and `enterprise_uuid` viewed
        by an admin of that enterprise.

        Parameters:
        - `lms_user_id` (str): Filter results by the LMS user ID.
        - `enterprise_uuid` (str): Filter results by the Enterprise UUID.
        """
        lms_user_id = request.query_params.get('lms_user_id')
        enterprise_uuid = request.query_params.get('enterprise_uuid')
        if not lms_user_id or not enterprise_uuid:
            return Response(
                {"error": "Both 'lms_user_id' and 'enterprise_uuid' are required parameters."},
                status=HTTP_400_BAD_REQUEST
            )
        enterprise_customer_user = get_object_or_404(
            models.EnterpriseCustomerUser,
            user_id=lms_user_id,
            enterprise_customer__uuid=enterprise_uuid
        )
        enterprise_enrollments = models.EnterpriseCourseEnrollment.objects.filter(
            enterprise_customer_user=enterprise_customer_user
        )
        filtered_enterprise_enrollments = [record for record in enterprise_enrollments if record.course_enrollment]
        course_overviews = get_course_overviews([record.course_id for record in filtered_enterprise_enrollments])
        serialized_data = serializers.EnterpriseCourseEnrollmentAdminViewSerializer(
            filtered_enterprise_enrollments,
            many=True,
            context={
                'request': request,
                'enterprise_customer_user': enterprise_customer_user,
                'course_overviews': course_overviews,
            }
        ).data
        page = self.paginate_queryset(serialized_data)
        grouped_data = self._group_course_enrollments_by_status(page)
        return self.get_paginated_response(grouped_data)

    def _group_course_enrollments_by_status(self, course_enrollments):
        """
        Groups course enrollments by their status.

        Args:
            enrollments (list): List of course enrollment dictionaries.

        Returns:
            dict: A dictionary where keys are status names and values are lists of enrollments with that status.
        """
        statuses = {
            CourseRunProgressStatuses.IN_PROGRESS: [],
            CourseRunProgressStatuses.UPCOMING: [],
            CourseRunProgressStatuses.COMPLETED: [],
            CourseRunProgressStatuses.SAVED_FOR_LATER: [],
        }
        for enrollment in course_enrollments:
            status = enrollment.get('course_run_status')
            if status in statuses:
                statuses[status].append(enrollment)
        return statuses
